from typing import Coroutine


class _ContextManager(Coroutine):
    """Define asynchronous context manager object. Implement generator methods."""

    __slots__ = ('_coro', '_obj')

    def __init__(self, coro):
        self._coro = coro
        self._obj = None

    def send(self, value):
        return self._coro.send(value)

    def throw(self, typ, val=None, tb=None):
        if val is None:
            return self._coro.throw(typ)
        elif tb is None:
            return self._coro.throw(typ, val)
        else:
            return self._coro.throw(typ, val, tb)

    def close(self):
        return self._coro.close()

    @property
    def gi_frame(self):
        return self._coro.gi_frame

    @property
    def gi_running(self):
        return self._coro.gi_running

    @property
    def gi_code(self):
        return self._coro.gi_code

    def __next__(self):
        return self.send(None)

    def __iter__(self):
        return self._coro.__await__()

    def __await__(self):
        return self._coro.__await__()

    async def __aenter__(self):
        self._obj = await self._coro
        return self._obj

    async def __aexit__(self, exc_type, exc, tb):
        await self._obj.close()
        self._obj = None


class _ClientContextManager(_ContextManager):
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Override aexit dunder method in _ContextManager. A client should shutdown forcibly when an exception is
        caught."""
        if exc_type is not None:
            await self._obj.close()
        else:
            await self._obj.shutdown()
        self._obj = None
